#ifndef LNM_HTTP_REQ
#define LNM_HTTP_REQ

#include <regex.h>
#include <stdint.h>
#include <stdlib.h>

#include "picohttpparser.h"

#include "lnm/common.h"
#include "lnm/http/consts.h"
#include "lnm/http/route.h"

#define LNM_HTTP_MAX_REQ_HEADERS 32
#define LNM_HTTP_MAX_REGEX_GROUPS 4

/**
 * Internal representation of a header in a request, defined using offsets
 * relative to the full buffer.
 */
typedef struct lnm_http_req_ihdr {
  struct {
    size_t o;
    size_t len;
  } name;
  struct {
    size_t o;
    size_t len;
  } value;
} lnm_http_req_ihhr;

/**
 * Represents the parsed HTTP request
 */
typedef struct lnm_http_req {
  struct {
    char *s;
    size_t len;
    bool owned;
  } buf;
  int minor_version;
  lnm_http_method method;
  lnm_http_route_match route_match;
  struct {
    size_t o;
    size_t len;
  } path;
  struct {
    size_t o;
    size_t len;
  } query;
  struct {
    lnm_http_req_ihhr arr[LNM_HTTP_MAX_REQ_HEADERS];
    size_t len;
  } headers;
  struct {
    uint64_t expected_len;
    uint64_t len;
    char *buf;
    bool owned;
  } body;
} lnm_http_req;

typedef enum lnm_http_parse_err {
  lnm_http_parse_err_ok = 0,
  lnm_http_parse_err_incomplete,
  lnm_http_parse_err_invalid,
  lnm_http_parse_err_unknown_method,
} lnm_http_parse_err;

/**
 * Try to parse the given buffer into an HTTP request.
 *
 * @param req request to store parsed data in
 * @param buf buffer to parse; might be modified
 * @param len length of buf
 */
lnm_http_parse_err lnm_http_req_parse(lnm_http_req *req, char *buf, size_t len);

/**
 * Reset the given request object, free'ing all its relevant parts and allowing
 * it to be reused as a new object.
 *
 * @param req object to reset
 */
void lnm_http_req_reset(lnm_http_req *req);

/**
 * Represents an actual header value, with offsets already resolved.
 */
typedef struct lnm_http_req_hdr {
  struct {
    const char *s;
    size_t len;
  } name;
  struct {
    const char *s;
    size_t len;
  } value;
} lnm_http_req_hdr;

/**
 * Retrieve a known type header from the request.
 *
 * Pointers retrieved from this function should never be used between step
 * functions; simply request the header again if you need to.
 *
 * @param out where to store pointer to struct representing header
 * @param req request to look for header in
 * @param type type of header to look for
 */
lnm_err lnm_http_req_header_get(lnm_http_req_hdr *out, lnm_http_req *req,
                                lnm_http_header type);

/**
 * Retrieve a header from the request using a case-insensitive name.
 *
 * Pointers retrieved from this function should never be used between step
 * functions; simply request the header again if you need to.
 *
 * @param out where to store pointer to struct representing header
 * @param req request to look for header in
 * @param name name of the header; matches case-insensitive
 */
lnm_err lnm_http_req_header_get_s(lnm_http_req_hdr *out, lnm_http_req *req,
                                  const char *name);

/**
 * Retrieve a named key segment from the matched route.
 *
 * @param out where to write pointer to string
 * @param key key of the segment
 * @return length of the outputted char buffer, or 0 if the key doesn't exist
 */
size_t lnm_http_req_route_segment(const char **out, lnm_http_req *req,
                                  const char *key);

/**
 * Retrieve a specific key-value parameter from a header.
 *
 * Pointers retrieved from this function should never be used between step
 * functions; simply request the header again if you need to.
 *
 * @param out output struct
 * @param header header to look in
 * @param key name of the parameter to return from the header
 */
size_t lnm_http_req_header_param(const char **out,
                                 const lnm_http_req_hdr *header,
                                 const char *key);

#endif
